=begin
(c) TIG 2010

raytrace.rb

Place the script in the Plugins Folder.

Usage: 'RayTrace'
       Type raytrace in the Ruby Console or 
       pick 'RayTrace' off the Plugins Menu.
       Follow the prompts:
       Pick the Start Point of the 'Ray Trace'.
       A guide shows the possible 'Ray' path.
       Pick the End point of the 'Ray' on any face.
       A 'Ray' is drawn between the two picked points as a guide-line with 
       guide-points.
       The reflected 'Ray' is then drawn - it extends until it hits a 
       face.  If it hits something then a guide-point is drawn, 
       otherwise the reflected 'Ray' guide-line is drawn infinitely long.
       If the 'Ray' has hit a face then a dialog asks if you want to 
       'Bounce the Ray': answer Yes to make the next part of the 'Ray' 
       or No to stop this 'Ray'.
       If these is no hit then that 'Ray' set automatically ends.
       The tool now repeats: to stop select another tool or press <Esc>.
       The set of 'Rays' is grouped for ease of identification or later 
       deletion.
       Clear Guide-points and Lines using third party-scripts or use 
       Edit > Delete Guides...
       
       'RaySpray'
       Type rayspray in the Ruby Console or 
       pick 'RaySpray' off the Plugins Menu.
       Follow the prompts:
       Pick the start point of the 'Ray Spray'.
       A dialog asks if you want to constrain the spray in X/Y/Z-planes 
       or 3D.
       It also asks for the angle of separation of each of the spray's 
       rays - default is 30 degrees - closely angled rays take more time 
       to do - 3D rays are obviously more intensive.
       The 'Ray' guide-lines and points are grouped, as in 'RayTrace' .      
       There is no 'bounce' option with this tool, but once it the 
       'Ray Spray' has been made a 'Ray Trace' can be done on any 
       desired Ray's points, with bounces as required...
       After one use the tool does NOT repeat, it needs reactivating...
       
Donations:
        Are welcome [by PayPal], please use 'TIGdonations.htm' in the 
        ../Plugins/TIGtools/ folder - or PM TIG for details...
       
Version:
        1.0 20100324 First issue.
        1.1 20100325 Possible Bounces added.  Ray Spray added.
        1.2 20100326 Bouncings now stops on 'No' !  Inferencing improved.

=end

require 'sketchup.rb'

class Raytrace ###################################

def initialize
  @ip=nil
  @ip1=nil
  @ip2=nil
  @p1=nil
  @p2=nil
  @p3=nil
  @face=nil
  @group=nil
end

def activate()
  @ip=Sketchup::InputPoint.new
  @ip1=Sketchup::InputPoint.new
  @ip2=Sketchup::InputPoint.new
  @p1=[]
  @p2=[]
  @p3=[]
  @face=nil
  @group=nil
  @model=Sketchup.active_model
  @ents=@model.active_entities
  @msg="RayTrace: Pick Start Point..."
  @state=1
end

def deactivate(view)
  @ip=nil 
  @ip1=nil
  @ip2=nil
  @p1=[]
  @p2=[]
  @p3=[]
  @face=nil
  @group=nil
  Sketchup::set_status_text("")
  view.invalidate
  Sketchup.send_action("selectSelectionTool:")
  return nil
end

def onCancel(reason,view)
  @group.erase! if @group and @group.valid?
  view.invalidate
  self.deactivate(view)
end

def resume(view)
  Sketchup::set_status_text(@msg)
  self.draw(view)
end

def enableVCB?
  return true
end

def onMouseMove(flags,x,y,view)
  @ip.pick(view,x,y)
  view.lock_inference
  @ip.draw(view)
  view.tooltip=("RayTrace: ")+@ip.tooltip
  case @state
   when 1
    view.guess_target
    if @ip.valid? and @ip!=@ip1
      @ip1.copy!(@ip)
      @ip1.draw(view)
      view.invalidate
    end#if
   when 2
    view.guess_target
    @msg=("RayTrace: Pick End Point...")
    @ip2.pick(view,x,y,@ip1)
    if @ip2.valid? and @ip2!=@ip1
      @ip2.draw(view)
      view.invalidate
    end#if
   else
    view.invalidate
  end#case
  Sketchup::set_status_text(@msg)
  Sketchup.set_status_text((""),SB_VCB_LABEL)
  Sketchup.set_status_text("",SB_VCB_VALUE)
end#def

def onKeyDown(key,repeat,flags,view)
    if key==CONSTRAIN_MODIFIER_KEY and repeat==1
        @shift_down_time=Time.now
        if view.inference_locked?
            view.lock_inference
        elsif @state==1 and @ip1.valid?
            view.lock_inference @ip1
        elsif @state==2 and @ip2.valid?
            view.lock_inference @ip2, @ip1
        end
    end
end#def

def onKeyUp(key, repeat, flags, view)
    if key==CONSTRAIN_MODIFIER_KEY and view.inference_locked? and (Time.now - @shift_down_time) > 0.5
        view.lock_inference
    end
    view.invalidate
end#def

def onLButtonDown(flags,x,y,view)
  if @state==1 ###start
    @ip1.pick(view,x,y)
    if @ip1.valid?
      @p1=@ip1.position
      @ip.clear
      @state=2
    end#if
  elsif @state==2 ### second point for face bounce
    ph=view.pick_helper
    ph.do_pick(x,y)
    @face=ph.picked_face
    @ip2.pick(view,x,y)
    if @face and @ip2.valid? and @ip2!=@ip1
      @p2=@ip2.position
      view.invalidate
      @state=3
      self.create_geometry(view)
    end#if
  end#if
end#def

def onLButtonUp(flags,x,y,view)
  view.invalidate if view
end

def draw(view)
  ###
  @ip.draw(view)if @ip and @ip.valid?
  @ip1.draw(view)if @ip1 and @ip1.valid?
  @ip2.draw(view)if @ip2 and @ip2.valid?
  case @state
    when 1
     p=@ip.position
     view.draw_points([p],7,5)
     @ip1.draw(view)if @ip1 and @ip1.valid?
    when 2
     p1=@ip1.position
     view.draw_points([p1],7,5)
     p2=@ip2.position
     view.draw_points([p2],7,5)
     view.draw_line(@p1,@ip2.position)if @p1 and @ip2.valid?
  end#case
  ###
end

def create_geometry(view)
  @model.start_operation("RayTrace")
  ### make group
  @group=@ents.add_group
  gents=@group.entities
  ### draw first ray
  gents.add_cline(@p1,@p2)
  gents.add_cpoint(@p1)
  gents.add_cpoint(@p2)
  ### calculate reflected ray's vector
  vector1=@p1.vector_to(@p2)
  normal=@face.normal
  angle=vector1.angle_between(normal)
  perp=vector1.cross(normal)
  if perp.length !=0
    pt=@p2.offset(normal,1)
    tr=Geom::Transformation.rotation(@p2,perp,angle-180.degrees)
    pt.transform!(tr)
    vector2=@p2.vector_to(pt)
    rayt=@model.raytest([@p2,vector2])
    @p3=nil
    @p3=rayt[0]if rayt
    cl=gents.add_cline(@p2,pt)
    gents.add_cpoint(@p3)if @p3
    cl.end=@p3
  end#if
  ### bounce ?
  if @p3
    bouncing=true
    p0=@p2
    ps=@p3
    hit=nil
    hit=rayt[1][-1]if rayt
    if hit.class==Sketchup::Face
      face=hit
    elsif hit.class==Sketchup::Edge
      face=hit.faces[0]
    else
      bouncing=false
    end#if
    ###
    while bouncing
      bounce=UI.messagebox("RayTrace: Bounce the Ray ?",MB_YESNO,"")###6=YES 7=NO
      if bounce==6 and face
        vector1=p0.vector_to(ps)
        normal=face.normal
        angle=vector1.angle_between(normal)
        perp=vector1.cross(normal)
        pt=ps.offset(normal,1)
        tr=Geom::Transformation.rotation(ps,perp,angle-180.degrees)
        pt.transform!(tr)
        vector2=ps.vector_to(pt)
        rayt=@model.raytest([ps,vector2])
        pe=nil
        pe=rayt[0]if rayt
        cl=gents.add_cline(ps,pt)
        gents.add_cpoint(pe)if pe
        cl.end=pe
        p0=ps
        ps=pe
        hit=nil
        hit=rayt[1][-1]if rayt
        if hit.class==Sketchup::Face
          face=hit
        elsif hit.class==Sketchup::Edge
          face=hit.faces[0]
        else
          bouncing=false
        end#if
      else
        bouncing=false
      end#if
    end#while
  end#if
  ###
  @model.commit_operation
  ###
  self.activate()
  return nil
end

end#class ######################################

### shortcut defn
def raytrace()
  Sketchup.active_model.select_tool(Raytrace.new)
end#def


###~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class Rayspray ###################################

def initialize
  @ip=nil
  @ip1=nil
  @p1=nil
  @p2=nil
  @p3=nil
  @group=nil
end

def activate()
  @ip=Sketchup::InputPoint.new
  @ip1=Sketchup::InputPoint.new
  @p1=[]
  @p2=[]
  @p3=[]
  @group=nil
  @model=Sketchup.active_model
  @ents=@model.active_entities
  @msg="RaySpray: Pick Start Point..."
  @state=1
end

def deactivate(view)
  @ip=nil 
  @ip1=nil
  @p1=[]
  @p2=[]
  @p3=[]
  @group=nil
  Sketchup::set_status_text("")
  view.invalidate
  Sketchup.send_action("selectSelectionTool:")
  return nil
end

def onCancel(reason,view)
  @group.erase! if @group and @group.valid?
  view.invalidate
  self.deactivate(view)
end

def resume(view)
  Sketchup::set_status_text(@msg)
  self.draw(view)
end

def enableVCB?
  return true
end

def onMouseMove(flags,x,y,view)
  @ip.pick(view,x,y)
  view.lock_inference
  view.tooltip=("RaySpray: ")+@ip.tooltip
  case @state
   when 1
    if @ip.valid? and @ip!=@ip1
      @ip1.copy!(@ip)
      view.invalidate
    end#if
   when 2
    @msg=("RaySpray: Settings...")
   else
    view.invalidate
  end#case
  Sketchup::set_status_text(@msg)
  Sketchup.set_status_text((""),SB_VCB_LABEL)
  Sketchup.set_status_text("",SB_VCB_VALUE)
end#def

def onKeyDown(key,repeat,flags,view)
    if key==CONSTRAIN_MODIFIER_KEY and repeat==1
        @shift_down_time=Time.now
        if view.inference_locked?
            view.lock_inference
        elsif @state==1 and @ip1.valid?
            view.lock_inference @ip1
        end
    end
end#def

def onKeyUp(key, repeat, flags, view)
    if key==CONSTRAIN_MODIFIER_KEY and view.inference_locked? and (Time.now - @shift_down_time) > 0.5
        view.lock_inference
    end
    view.invalidate
end#def

def onLButtonDown(flags,x,y,view)
  if @state==1 ###start
    @ip1.pick(view,x,y)
    if @ip1.valid?
      @p1=@ip1.position
      @ip.clear
      @state=2
      self.deactivate(view) if not self.dialog(view)
      self.create_geometry(view)
    end#if
  end#if
end#def

def onLButtonUp(flags,x,y,view)
  view.invalidate if view
end

def draw(view)
  ###
end

def dialog(view)
  @con="3D"
  @ang=30.0
  a1="Constraint: "
  a2="Angle: "
  prompts=[a1,a2]
  values=[@con, @ang]
  xyz=["X-plane|Y-plane|Z-plane|3D"]
  popups=[xyz,""]
  title="Ray Spray Settings"
  results=UI.inputbox(prompts,values,popups,title)
  if not results
    self.deactivate(view) 
    return nil
  end#if
  @con=results[0].to_s
  @ang=results[1].to_f
  ###
  return true
end

def create_geometry(view)
  @model.start_operation("RaySpray")
  ### make group
  @group=@ents.add_group
  @gents=@group.entities
  ### draw first pt
  @gents.add_cpoint(@p1)
  ###
  case @con
    when "X-plane"
      vector=Geom::Vector3d.new(1,0,0)
      normal=Geom::Vector3d.new(0,1,0)
      self.one_bounce(view,vector,normal)
    when "Y-plane"
      vector=Geom::Vector3d.new(0,1,0)
      normal=Geom::Vector3d.new(1,0,0)
      self.one_bounce(view,vector,normal)
    when "Z-plane"
      vector=Geom::Vector3d.new(1,0,0)
      normal=Geom::Vector3d.new(0,0,1)
      self.one_bounce(view,vector,normal)
    else ### 3D
      vector=Geom::Vector3d.new(1,0,0)
      normal=Geom::Vector3d.new(0,0,1)
      self.one_bounce(view,vector,normal,true)
  end#case
  ###
  @model.commit_operation
  ###
  self.deactivate(view)
  return nil
end

def one_bounce(view,vector,normal,in3D=nil)
  ### puts @con ###########
  if not in3D
    count=(360.0/@ang).to_i
    count.times do |i|
      ang=@ang*i
      tr=Geom::Transformation.rotation(@p1,normal,ang.degrees)
      vec=vector.transform(tr)
      rayt=@model.raytest([@p1,vec])
      pt=@p1.offset(vec,1)
      @p2=nil
      @p2=rayt[0]if rayt
      @gents.add_cpoint(@p2)if @p2
      cl=@gents.add_cline(@p1,pt)
      cl.end=@p2
      ### check hit a face
      hit=nil
      hit=rayt[1][-1]if rayt
      if hit.class==Sketchup::Face
        face=hit
      elsif hit.class==Sketchup::Edge
        face=hit.faces[0]
      else
        next
      end#if
      ###
      if @p2 ### reflect it just once
        ### calculate reflected ray's vector
        vector1=@p1.vector_to(@p2)
        fnormal=face.normal
        angle=vector1.angle_between(fnormal)
        perp=vector1.cross(fnormal)
        pt=@p2.offset(fnormal,1)
        if perp.length !=0
          tr=Geom::Transformation.rotation(@p2,perp,angle-180.degrees)
          pt.transform!(tr)
          vector2=@p2.vector_to(pt)
          rayt=@model.raytest([@p2,vector2])
          @p3=nil
          @p3=rayt[0]if rayt
          cl=@gents.add_cline(@p2,pt)
          @gents.add_cpoint(@p3)if @p3
          cl.end=@p3
        end#if
      end
    end#times
  else ### it's 3d
    self.one_bounce(view,vector,normal)### make z-plane version
    count=(360.0/@ang).to_i
    count.times do |i|
      ang=@ang*i
      vec=vector
      tr=Geom::Transformation.rotation(@p1,vec,ang.degrees)
      nor=normal.transform(tr)
      self.one_bounce(view,vec,nor)
    end#times
  end#if
end

end#class ######################################

### shortcut defn
def rayspray()
  Sketchup.active_model.select_tool(Rayspray.new)
end#def


###~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

### menu
if not file_loaded?(__FILE__)
  textstring="RayTrace"
  instructions=": Pick Start Point, then End Point on a Face: If could 'Bounce' choose Yes/No..."
  cmd=UI::Command.new(textstring){raytrace()}
  UI.menu("Plugins").add_item(textstring){raytrace()}
  cmd.status_bar_text=textstring+instructions
  textstring1="RaySpray"
  instructions1=": Pick Start Point, then Constraints [X/Y/Z-plane or 3D] and Angle..."
  cmd=UI::Command.new(textstring1){rayspray()}
  UI.menu("Plugins").add_item(textstring1){rayspray()}
  cmd.status_bar_text=textstring1+instructions1
end
#-----------------------------------------------------------------------------
file_loaded(__FILE__)
###

